#!/usr/bin/env python3

# Command line interface to start crawlers

import argparse
import os
import signal
import sys
import threading
import validators
from crawler import StreamCrawler

import crawler.trafikverket as trv
import crawler.trafikverket.railway as railway

VERSION = "1.3.0"

parser = argparse.ArgumentParser(description="Crawler for Trafikverket's Trafikinfo API - version " + VERSION)

# Add the arguments
parser.add_argument('objecttype',
                       metavar='objecttype',
                       type=str,
                       help='the API objecttype to crawl',
                       choices=["TrainAnnouncement", "TrainMessage"])

parser.add_argument('identifier',
                        metavar="identifier",
                        type=str,
                        help='the identifier to crawl data for (for railway this is the station location signature "trafikplatssignatur")',
                        )

parser.add_argument('-s', '--stream', 
                        action='store_true',
                        help='use streaming in real-time instead of daily crawling'
                        )

parser.add_argument('-f', '--force', 
                        action='store_true',
                        help='force new connection instead of reconnecting'
                        )

parser.add_argument('-o', '--output', 
                metavar="path",
                type=str,
                help='the output folder',
                default=None
                )

parser.add_argument("-w", '--webhook',
                metavar="webhookurl",
                type=str,
                help="Webhook URL to trigger when warnings or exceptions occur",
                default=None,
                required=False)

parser.add_argument("-a", '--authkey',
                metavar="authkey",
                type=str,
                help="The API authkey to use as obtained from https://data.trafikverket.se/get-started",
                required=True)

parser.add_argument("-v", '--verbose',
                action='store_true',
                help="Prints extra debug information")

parser.add_argument("-m", '--mute',
                action='store_true',
                help="Mute output of warnings to webhook.")

parser.add_argument("-q", '--query',
                action='store_true',
                help="Print query only instead of starting a crawler")

# Validate arguments
args = parser.parse_args()

if args.output is None:
    #not specified, use default
    if args.stream:
        args.output = os.path.join(".", "stream", args.objecttype, args.identifier)
    else:
        args.output = os.path.join(".", "daily", args.objecttype, args.identifier)
else:
    #specified, check if exists
    if not os.path.isdir(args.output):
        try:
            os.makedirs(args.output)
            print("The output path {output} has been created.".format(output=args.output))
        except Exception as e:
            print("The output path {output} could not be created: {exception}".format(output=args.output, exception=e))
            exit(-1)

if not args.webhook is None and not validators.url(args.webhook):
    print("Webhook is not a valid URL.")
    exit(-1)

if args.force and not args.stream:
    print("Forcing a new connection is only available when streaming")
    exit(-1)

# Prepare crawler
crawler_args = {
    "path": args.output,
    "authkey": args.authkey, 
    "identifier": args.identifier,
    "version": VERSION
}

if args.objecttype == "TrainAnnouncement" and args.stream:
    c = railway.StreamTrainAnnouncementCrawler(**crawler_args)
elif args.objecttype == "TrainAnnouncement" and not args.stream:
    c = railway.DailyTrainAnnouncementCrawler(**crawler_args)
elif args.objecttype == "TrainMessage" and args.stream:
    c = railway.StreamTrainMessageCrawler(**crawler_args)
elif args.objecttype == "TrainMessage" and not args.stream:
    c = railway.DailyTrainMessageCrawler(**crawler_args)

if not args.webhook is None:
    c.message_webhook_url = args.webhook

if args.verbose:
    c.DEBUG = True
    print("Verbose output enabled.")

if args.mute:
    c.MUTE = True
    print("Only exceptions are sent to webhook. Warnings are shown on console and written to the logfile.")

if args.query:
    print(c._get_query())
    sys.exit(0)

# Print status
def print_status():
    if not c.messages.empty():
        print("")
        while not c.messages.empty():
            print(c.messages.get())


    if not c.stop_request.is_set() and isinstance(c, StreamCrawler) and not c.connected.is_set():
        print("{name}\tconnecting\t{warnings} warnings\t{exceptions} exceptions".format(
            name=c.path, writes=c.writes, warnings=c.warnings, exceptions=c.exceptions),end='\r')
    elif not c.stop_request.is_set() and c.running.is_set():
        print("{name}\t{writes} writes\t{warnings} warnings\t{exceptions} exceptions".format(
            name=c.path, writes=c.writes, warnings=c.warnings, exceptions=c.exceptions),end='\r')
    elif c.stop_request.is_set() and c.running.is_set():
        print("{name}\tstopping\t{warnings} warnings\t{exceptions} exceptions".format(
            name=c.path, writes=c.writes, warnings=c.warnings, exceptions=c.exceptions),end='\r')   
    else:
        print("{name}\tnot running\t{warnings} warnings\t{exceptions} exceptions".format(
            name=c.path, writes=c.writes, warnings=c.warnings, exceptions=c.exceptions),end='\r')

    if c.running.is_set() or not c.stop_request.is_set():
        threading.Timer(1, print_status).start()
    else:
        sys.exit(0)

print_status()

#Log command run
c.log_status("Command run: {cmd}".format(cmd=" ".join(sys.argv)))

# Start crawling
if args.stream and args.force:
    c.start(force_new_connection=True)
else:
    c.start()

def signal_handler(sig, frame):
    c.messages.put("Crawler stopping...")
    c.stop()
    if c.join(wait_seconds=3):
        print("\nCrawler stopped.")
    sys.exit(0)

signal.signal(signal.SIGINT, signal_handler)
c.messages.put("∞ Crawler running. Press CTRL-C to stop.")
