import datetime
import crawler
import pandas as pd
import os

#=====================================================================
# Define constants
#=====================================================================

ENDPOINT_URL = "https://api.trafikinfo.trafikverket.se/v2/data.json"

#=====================================================================
# Trafikverket API Query building
#=====================================================================

def format_datetime(dt, utc=False):
    """Format a datetime object as in Trafikverkets API

    Args:
        dt (datetime.datetime): The datetime object to format
        utc (bool, optional): Use UTC format instead of local time. Defaults to False (local time).

    Returns:
        str: The formatted string
    """

    if utc:
        return dt.astimezone(tz=datetime.timezone.utc).isoformat(timespec='seconds').replace("+00:00","Z")
    else:
        return dt.astimezone().isoformat(timespec='seconds')

def build_query(authkey, attributes = {}, filter = "", columns = []):
    attributestr = " ".join(["{key}='{value}'".format(key=key, value=value) for key, value in attributes.items()])    
    includestr = "\n".join(["<INCLUDE>{column}</INCLUDE>".format(column=column) for column in columns])

    query = """
    <REQUEST>
        <LOGIN authenticationkey='{authkey}' />
        <QUERY {attributestr}>
            <FILTER>
                {filter}
            </FILTER>
            {includestr}
        </QUERY>
    </REQUEST>
    """.format(authkey=authkey, attributestr = attributestr, filter=filter, includestr=includestr)

    return query

#=====================================================================
# TrafikverketDailyCrawler
#=====================================================================

class TrafikverketDailyCrawler (crawler.PeriodicCrawler):
    """Crawler that daily fetches data from Trafikverkets API
    """

    def __init__(self, path, authkey, objecttype, columns, identifier="", interval=60*60, version="1.0", output_version=1):
        """Initializes the crawler

        Args:
            path (str): The path where the crawler writes the data
            authkey (str): Authkey for Trafikverkets API to use for requests, obtained from https://data.trafikverket.se/get-started
            objecttype (str): The objecttype to crawl as in the TrV API (e.g. TrainAnnouncement)
            columns (list): A list of string giving the column names to include
            identifier (str, optional): The identifier of the data object to monitor (e.g. railway station location signature). Defaults to "".
            interval (float, optional): Interval in seconds that the crawler checks if data needs to be crawled. Defaults to 60*60.
            version (str): Identifies the version of the crawler code
            output_version (int): Gives the version of the output format. Increment when adding/chaning columns to crawl.
        """

        super().__init__(ENDPOINT_URL, path, authkey, interval=interval, version=version, output_version=output_version)
        self.objecttype = objecttype
        self.columns = columns
        self.identifier = identifier

    def _get_query(self):
        raise NotImplementedError("Missing query")
    
    def crawl(self):
        todayMidnight = datetime.datetime.now().replace(hour=0, minute=0, second=0, microsecond=0)
        yesterdayMidnight = todayMidnight - datetime.timedelta(days=1)
        date_str = yesterdayMidnight.strftime("%Y-%m-%d")
        filename = os.path.join(self.path, f"{self.objecttype}_{self.identifier}_{date_str}.csv")
        
        if not os.path.exists(filename):
            if self.DEBUG:
                self.log_status("Crawling yesterdays data...")

            #Fetch data for yesterday            
            try:
                data = self._attempt(lambda: self._query(self._get_query()), description="Querying yesterdays data")

                #check for error message
                try:
                    self.log_warning(data["RESPONSE"]["RESULT"][0]["ERROR"])
                except Exception:
                    pass

                #get data
                try:
                    data =  pd.DataFrame(data["RESPONSE"]["RESULT"][0][self.objecttype], columns = self.columns)
                    data.insert(loc=0, column="TimeReceived", value=len(data)*[format_datetime(datetime.datetime.now(), utc=True)])

                    if self.DEBUG:
                        self.log_info("Received {rows} rows.".format(rows=len(data)))
                except Exception as e:
                    self.log_exception(e)

                #write to csv
                with self.write_lock:
                    data.to_csv(filename, index=False)
                    self.writes += 1
                    
                if self.DEBUG:
                    self.log_info("Written {rows} rows.".format(rows=len(data)))
            except Exception as e:
                self.log_exception(e)
        else:
            if self.DEBUG:
                self.log_info("Latest daily data already written. No need to crawl at this point.")


#=====================================================================
# TrafikverketStreamCrawler
#=====================================================================

class TrafikverketStreamCrawler (crawler.StreamCrawler):
    """Crawler that continously crawls changed data using streaming from Trafikverkets API
    """
    def __init__(self, path, authkey, objecttype, columns, identifier="", version="1.0", output_version=1):
        """Initializes the crawler

        Args:
            path (str): The path where the crawler writes the data
            authkey (str): Authkey for Trafikverkets API to use for requests, obtained from https://data.trafikverket.se/get-started
            objecttype (str): The objecttype to crawl as in the TrV API (e.g. TrainAnnouncement)
            columns (list): A list of string giving the column names to include
            identifier (str, optional): The identifier of the data object to monitor (e.g. railway station location signature). Defaults to "".
            version (str): Identifies the version of the crawler code
            output_version (int): Gives the version of the output format. Increment when adding/chaning columns to crawl.
        """
        super().__init__(ENDPOINT_URL, path, authkey, version=version, output_version=output_version)
        self.objecttype = objecttype
        self.columns = columns
        self.identifier = identifier
        self.inactivity_seconds = 10*60*60 #restart connection if no events received for 10 hours

    @property
    def fileprefix(self):
        return f"{self.objecttype}_{self.identifier}_"

    def _get_query(self):
        raise NotImplementedError("Missing query")
    
    def _handle_events(self, data):
        #check for error message
        try:
            ex = Exception(data["RESPONSE"]["RESULT"][0]["ERROR"])
            self.log_exception(ex)
        except Exception:
            pass

        #get data
        try:
            data =  pd.DataFrame(data["RESPONSE"]["RESULT"][0][self.objecttype], columns = self.columns)
            data.insert(loc=0, column="TimeReceived", value=len(data)*[format_datetime(datetime.datetime.now(), utc=True)])
            
            if self.DEBUG:
                self.log_info("Received {rows} rows.".format(rows=len(data)))

            #write to csv
            with self.write_lock:
                data.to_csv(self.csvfile, index=False, mode='a', header=(not os.path.exists(self.csvfile)))
                self.writes += 1

            if self.DEBUG:
                self.log_info("Written {rows} rows.".format(rows=len(data)))
        except Exception as e:
            self.log_exception(e)