import datetime
import crawler.trafikverket as trv

#=====================================================================
# Train announcement query building
#=====================================================================

TRAIN_ANNOUNCEMENT_OUTPUT_VERSION = 3 #increase this after changing the query to make sure new output file is created

COLUMNS_TRAINANNOUNCEMENT = [   
    "ActivityId",
    "ActivityType",
    "Advertised",
    "AdvertisedTimeAtLocation",
    "AdvertisedTrainIdent",
    "Booking",
    "Canceled",
    "Deleted",
    "Deviation",
    "EstimatedTimeAtLocation",
    "EstimatedTimeIsPreliminary",
    "FromLocation",
    "InformationOwner",
    "LocationSignature",
    "MobileWebLink",
    "ModifiedTime",
    "NewEquipment",
    "Operator",
    "OtherInformation",
    "PlannedEstimatedTimeAtLocation",
    "PlannedEstimatedTimeAtLocationIsValid",
    "ProductInformation",
    "ScheduledDepartureDateTime",
    "Service",
    "LocationDateTimeOTN",
    "OperationalTrainNumber",
    "TimeAtLocation",
    "TimeAtLocationWithSeconds",
    "ToLocation",
    "TrackAtLocation",
    "TrainComposition",
    "TrainOwner",
    "TypeOfTraffic",
    "WebLink",
    "WebLinkName",
    "ViaFromLocation",
    "ViaToLocation",
    "DepartureDateOTN"
]

def get_train_announcement_query(authkey, startTime = None, endTime=None, locationSignature= "", sseurl=False, limit=None):
    attributes = {
        "objecttype": "TrainAnnouncement",
        "schemaversion": "1.8",
        "orderby": "TimeAtLocation",
        "includedeletedobjects": "true",
        "sseurl": sseurl
    }
    if not limit is None:
        attributes["limit"] = limit

    filter = """
        <EQ name="LocationSignature" value="{locationSignature}"/>
    """.format(locationSignature=locationSignature)

    if not startTime is None:
        filter += """
            <OR>
                <GTE name="AdvertisedTimeAtLocation" value="{startTime}" />
                <GTE name="TimeAtLocation" value="{startTime}" />
            </OR>
        """.format(startTime=trv.format_datetime(startTime))
    
    if not endTime is None:
        filter += """
            <OR>
                <LT name="AdvertisedTimeAtLocation" value="{endTime}" />
                <LT name="TimeAtLocation" value="{endTime}" />
            </OR>
        """.format(endTime=trv.format_datetime(endTime))
    
    return trv.build_query(authkey, attributes=attributes,filter=filter, columns= COLUMNS_TRAINANNOUNCEMENT)


#=====================================================================
# Train message query building
#=====================================================================

TRAIN_MESSAGE_OUTPUT_VERSION = 3 #increase this after changing the query to make sure new output file is created

COLUMNS_TRAINMESSAGE = [
    "CountyNo",
    "Deleted",
    "EndDateTime",
    "EventId",
    "ExternalDescription",
    "Geometry",
    "Header",
    "LastUpdateDateTime",
    "ModifiedTime",
    "PrognosticatedEndDateTimeTrafficImpact",
    "ReasonCode",
    "StartDateTime",
    "TrafficImpact"
]

def get_train_message_query(authkey, startTime = None, endTime=None, locationSignature= "", sseurl=False, limit=None):
    attributes = {
        "objecttype": "TrainMessage",
        "schemaversion": "1.7",
        "orderby": "StartDateTime",
        "includedeletedobjects": "true",
        "sseurl": sseurl
    }
    if not limit is None:
        attributes["limit"] = limit

    filter = """
        <IN name="TrafficImpact.AffectedLocation.LocationSignature" value="{locationSignature}"/>
    """.format(locationSignature=locationSignature)

    if not startTime is None:
        filter += """<GTE name="StartDateTime" value="{startTime}" />\n""".format(startTime=trv.format_datetime(startTime))

    if not endTime is None:
        filter += """<LT name="StartDateTime" value="{endTime}" />\n""".format(endTime=trv.format_datetime(endTime))

    return trv.build_query(authkey, attributes=attributes,filter=filter, columns= COLUMNS_TRAINMESSAGE)



#=====================================================================
# Daily Trafikverket crawlers
#=====================================================================

class DailyTrainAnnouncementCrawler (trv.TrafikverketDailyCrawler):
    """Crawler that daily fetches all train announcements for the previous day for a given location.
    """

    def __init__(self, path, authkey, interval=60*60, identifier="", version="1.0"):
        """Initializes the crawler

        Args:
            path (str): The path where the crawler writes the data
            authkey (str): Authkey for Trafikverkets API to use for requests, obtained from https://data.trafikverket.se/get-started
            interval (float, optional): Interval in seconds that the crawler checks if data needs to be crawled. Defaults to 60*60.
            identifier (str, optional): The locationSignature of the station to monitor. Defaults to "".
        """
        super().__init__(path, authkey, "TrainAnnouncement", COLUMNS_TRAINANNOUNCEMENT,
                         identifier=identifier, interval=interval, version=version, output_version=TRAIN_ANNOUNCEMENT_OUTPUT_VERSION)

    def _get_query(self):
        todayMidnight = datetime.datetime.now().replace(hour=0, minute=0, second=0, microsecond=0)
        yesterdayMidnight = todayMidnight - datetime.timedelta(days=1)
        return get_train_announcement_query(self.authkey, startTime=yesterdayMidnight, endTime=todayMidnight, locationSignature=self.identifier)


class DailyTrainMessageCrawler (trv.TrafikverketDailyCrawler):
    """Crawler that daily fetches all train messages (disruption messages) for the previous day for a given location.
    """

    def __init__(self, path, authkey, interval=60*60, identifier="", version="1.0"):
        """Initializes the crawler

        Args:
            path (str): The path where the crawler writes the data
            authkey (str): Authkey for Trafikverkets API to use for requests, obtained from https://data.trafikverket.se/get-started
            interval (float, optional): Interval in seconds that the crawler checks if data needs to be crawled. Defaults to 60*60.
            identifier (str, optional): The locationSignature of the station to monitor. Defaults to "".
        """
        super().__init__(path, authkey, "TrainMessage", COLUMNS_TRAINMESSAGE,
                         identifier=identifier, interval=interval, version=version, output_version=TRAIN_MESSAGE_OUTPUT_VERSION)

    def _get_query(self):
        todayMidnight = datetime.datetime.now().replace(hour=0, minute=0, second=0, microsecond=0)
        yesterdayMidnight = todayMidnight - datetime.timedelta(days=1)
        return get_train_message_query(self.authkey, startTime=yesterdayMidnight, endTime=todayMidnight, locationSignature=self.identifier)

#=====================================================================
# Streaming Trafikverket crawlers
#=====================================================================

class StreamTrainAnnouncementCrawler (trv.TrafikverketStreamCrawler):
    """Crawler that continously monitors all train announcements (arrival, departure) for a given location.
    """
    def __init__(self, path, authkey, identifier="", version="1.0"):
        """Initializes the crawler

        Args:
            path (str): The path where the crawler writes the data
            authkey (str): Authkey for Trafikverkets API to use for requests, obtained from https://data.trafikverket.se/get-started
            identifier (str, optional): The locationSignature of the station to monitor. Defaults to "".
        """
        super().__init__(path, authkey, "TrainAnnouncement", COLUMNS_TRAINANNOUNCEMENT,
                         identifier=identifier, version=version, output_version=TRAIN_ANNOUNCEMENT_OUTPUT_VERSION)

    def _get_query(self):
        return get_train_announcement_query(self.authkey, locationSignature=self.identifier, sseurl=True)


class StreamTrainMessageCrawler (trv.TrafikverketStreamCrawler):
    """Crawler that continously monitors all train messages (disruption messages) for a given location.
    """
    def __init__(self, path, authkey, identifier="", version="1.0"):
        """Initializes the crawler

        Args:
            path (str): The path where the crawler writes the data
            authkey (str): Authkey for Trafikverkets API to use for requests, obtained from https://data.trafikverket.se/get-started
            identifier (str, optional): The locationSignature of the station to monitor. Defaults to "".
        """
        super().__init__(path, authkey, "TrainMessage", COLUMNS_TRAINMESSAGE,
                         identifier=identifier, version=version, output_version=TRAIN_MESSAGE_OUTPUT_VERSION)

    def _get_query(self):
        return get_train_message_query(self.authkey, locationSignature=self.identifier, sseurl=True)